from .auxfunc import (discretization, sjoin, condentropy)

from .sentropies import *
from .jentropies import *
from .ternary import *
from .coordsentropic import entcoords
from .entropytriangle import entriangle, entriangle_list


'''

Module for plotting the ET

'''

import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import matplotlib.lines as mlines

from matplotlib.lines import Line2D
from sys import exit as exit
from .ternary import ternary_axes_subplot
from .coordsentropic import *
from random import choice

#Auxiliar functions

def get_cmap(number, cmet = "1"):

    """
    Used for generating a set of random colors in order to differenciate the coordinates of each variable in the diagram. 
    A series of color maps can be applied (It will be randomly selected)

    > colors_markers = get_cmap(len(df.index))

    Parameters
    ----------
    number : Number of colours (1 per variable)

    Returns
    ----------
    colors : Returns a set of colors for creating the scatter plot (matplotlib.colors.LinearSegmentedColormap)

    """
    if(cmet=="1"):
        #cmaps = list(['Greys', 'Purples', 'Blues', 'Greens', 'Oranges', 'Reds','YlOrBr', 'YlOrRd', 'OrRd', 'PuRd', 'RdPu', 'BuPu','GnBu', 'PuBu', 'YlGnBu', 'PuBuGn', 'BuGn', 'YlGn'])
        cmaps = list(['spring','summer','autumn','winter','cool','Wistia'])
        cmaps = list(['summer'])
    else:
        cmaps = list(['summer'])
    return plt.cm.get_cmap(choice(cmaps) , number)



def markers(n) :
    
    #filled_markers = ('o', '*','+', 'x','D') ; mk = list()
    filled_markers = ('o') ; mk = list()
    for i in range(n) : mk.append(choice(filled_markers))
    return mk

    
    
def entriangle (edf,names = None, scale = 100 , fonts = 30 ,s_mk = 200 
                     ,gridl = 20 , ticks_size = 15 , pltscale = 20, 
                     chart_title = ""):
    
    """
    Function for creating and plotting points of the entropy triangle, independentlly of the type of triangle (CBET, SMET or CMET)

    > entropytriangle(edf,names = None, scale = 100 , fonts = 30 ,s_mk = 200 
                     ,gridl = 20 , ticks_size = 15 , pltscale = 20, 
                     chart_title = ""):

    Parameters
    ----------
    edf : Dataframe with the entropic measures calculated
    names : Used for plotting a list of Daframes (Defalut = NONE) 
            In case of list: Names of the dataframes (if empty, some names will be provided for the plotting)
    scale : Scale for the entropy triangle
    fonts : Fontsize of the Labels
    s_mk : Size of the marks for the triangle
    gridl : Space between each gridline
    tick_size: size of the numbers for the ticks
    pltscale : Scale of the matplotlib window
    chart_title: Plot´s title
    

    Returns
    ----------
    points : painted points 
    Shows the triangle
    
    """

    if(not isinstance(edf,pd.DataFrame)):
            exit("Can only work with Data Frames or lists of DataFrames! (df it´s not a DataFrame)")
               
    points = entcoords(edf, scale)

    #SMET PLOT
    if(hasSplitSmetCoords(edf) or hasAggregateSmetCoords(edf) or hasDualAggregateSmetCoords(edf)):

        figure, tax = ternary_axes_subplot.figure(scale=scale,angle = -60)
        figure.set_size_inches(pltscale, pltscale)
        tax.boundary(linewidth=2.0)
        tax.gridlines(multiple = gridl, color="blue")
    
        colors = get_cmap(len(edf.index)) ; names = list(edf.index)
        mk = markers(1)*len(edf)
        
        for i in range(len(edf.index)):
            tax.scatter(points[i:i+1], s = s_mk, marker = mk[i], color = colors(i), label = names[i] ,edgecolor='black', linewidth='0.1')

        if (hasSplitSmetCoords(edf)):

            if(chart_title != ""):
                tax.set_title(chart_title, fontsize = fonts + 5)
            else:
                tax.set_title("Source Multivariate split entropies (SMET)", fontsize = fonts + 5)
            tax.left_axis_label(r"$\Delta H'_{P_{X_i}}$", fontsize=fonts)
            tax.right_axis_label(r"$ M'_{P_{X_i}}$", fontsize=fonts)
            tax.bottom_axis_label(r"$ H'_{P_{X_i|X_i^c}}$", fontsize=fonts)

        elif (hasAggregateSmetCoords(edf)):

            if(chart_title != ""):
                tax.set_title(chart_title, fontsize = fonts + 5)
            else:
                tax.set_title("Source Multivariate entropies (SMET)", fontsize = fonts + 5)
            tax.left_axis_label(r"$\Delta H'_{Pi_{X}}$", fontsize=fonts)
            tax.right_axis_label(r"$ M'_{P_{X}}$", fontsize=fonts)
            tax.bottom_axis_label(r"$ VI'_{P_{X}}$", fontsize=fonts)

        elif (hasDualAggregateSmetCoords(edf)):

            if(chart_title != ""):
                tax.set_title(chart_title, fontsize = fonts + 5)
            else:
                tax.set_title("Dual Source Multivariate entropies (SMET)", fontsize = fonts + 5)
            tax.left_axis_label(r"$\Delta H'_{P_{X}}$", fontsize=fonts)
            tax.right_axis_label(r"$ D'_{P_{X}}$", fontsize=fonts)
            tax.bottom_axis_label(r"$ VI'_{P_{X}}$", fontsize=fonts)
    
    elif(hasCbetEntropicCoords(edf)):

        figure, tax = ternary_axes_subplot.figure(scale=scale)
        figure.set_size_inches(pltscale, pltscale)
        tax.boundary(linewidth=2.0)
        tax.gridlines(multiple = gridl, color="blue")

        mk = markers(1)*len(edf.index) ; colors = get_cmap(len(edf.index)) ; names = list(edf.index)
        
        for i in range(len(edf.index)):
            tax.scatter(points[i:i+1], s = s_mk , marker = mk[i], color = colors(i), label = names[i] ,edgecolor='black', linewidth='0.1')

        if(chart_title != ""):
            tax.set_title(chart_title, fontsize = fonts + 5)
        else:
            tax.set_title("Channel Bivariate entropy triangle (CBET)", fontsize = fonts + 5)
        tax.left_axis_label(r"$ VI_{P_{XY}}$", fontsize=fonts)
        tax.right_axis_label(r"$ MI_{P_{XY}}  $", fontsize=fonts)
        tax.bottom_axis_label(r"$\Delta H_{P_{X}\cdot{P_{Y}}}$", fontsize=fonts)


    #CMET PLOT
    elif(hasCmetEntropicCoords(edf)):

        figure, tax = ternary_axes_subplot.figure(scale=scale)
        figure.set_size_inches(pltscale, pltscale)
        tax.boundary(linewidth=2.0)
        tax.gridlines(multiple = gridl, color="blue")

        mk = markers(len(edf.index)) ; colors = get_cmap(len(edf.index)) ; names = list(edf.index)
        for i in range(len(edf.index)):
            tax.scatter(points[i:i+1], s = s_mk , marker = mk[i], color = colors(i), label = names[i] ,edgecolor='black', linewidth='0.3')

        if(chart_title != ""):
            tax.set_title(chart_title, fontsize = fonts + 5)
        else:
            tax.set_title("Channel Multivariate entropies (CMET)", fontsize = fonts + 5)
        tax.left_axis_label(r"$ VI'_{P_\overline{XY}}$", fontsize=fonts)
        tax.right_axis_label(r"$ 2\cdot{I'_{P_\overline{XY}}}  $", fontsize=fonts)
        tax.bottom_axis_label(r"$\Delta H'_{P_\overline{XY}}$", fontsize=fonts)

    else: 
        
        exit("Entropic coordinates needed for the diagram plotting")

    tax.ticks(axis='lbr', linewidth=1, multiple = gridl , fontsize = ticks_size )
    tax.clear_matplotlib_ticks()
    tax.legend(title = 'Features' ,labelspacing = 1.5 , fontsize = 10)
    
    tax.show()
    figure.savefig('plot.pdf')
    
    #return tax

    
    
def entriangle_list(edf, names = None, scale = 100 , fonts = 30 ,s_mk = 200 , gridl = 20 , ticks_size = 15 , pltscale = 17, chart_title = ""):
    
    """
    Function for creating and plotting points of the entropy triangle, independentlly of the type of triangle (CBET, SMET or CMET)
    This function just work with LISTS of dataframes

    > entropytriangle_list(edf,names = None, scale = 100 , fonts = 30 ,s_mk = 200 
                     ,gridl = 20 , ticks_size = 15 , pltscale = 20, 
                     chart_title = ""):

    Parameters
    ----------
    edf : list of entropy dataframes 
    names : Used for plotting a list of Daframes (Defalut = NONE) 
            In case of list: Names of the dataframes (if empty, some names will be provided for the plotting)
    scale : Scale for the entropy triangle
    fonts : Fontsize of the Labels
    s_mk : Size of the marks for the triangle
    gridl : Space between each gridline
    tick_size: size of the numbers for the ticks
    pltscale : Scale of the matplotlib window
    chart_title: Plot´s title
    

    Returns
    ----------
    points : painted points 
    Shows the triangle
    
    """

    if(not isinstance(edf, list)):
        exit('This function only works with LISTS of Dataframes, maybe you can try the etplot function for an individual DataFrame plot')
    
    if(not all(isinstance(edf[x],pd.DataFrame) for x in range(len(edf)))):
        exit('All values must be instances of DataFrames for entropy calculations')
    
    if(all(hasCbetEntropicCoords(edf[l]) for l in range(len(edf)))):
        
        figure, tax = ternary_axes_subplot.figure(scale=scale)
        figure.set_size_inches(pltscale, pltscale)
        tax.boundary(linewidth=2.0)
        tax.gridlines(multiple = gridl, color="blue")
        
        name = names
        colors = get_cmap(len(edf)); mk = markers(1)*len(edf) #mk = markers(1)*len(edf[i].index)
        for i in range(len(edf)):
            points = entcoords(edf[i], scale)
            #mk = markers(1)*len(edf[i].index) ; 
            name = str(names[i])
            for j in range(len(edf[i].index)):
                tax.scatter(points[j:j+1], s = s_mk , marker = mk[j], color = colors(i), label = name ,edgecolor='black', linewidth='0.3')

        if(chart_title != ""):
            tax.set_title(chart_title, fontsize = fonts + 5)
        else:
            tax.set_title("Channel Bivariate entropy triangle (CBET)", fontsize = fonts + 5)
        tax.left_axis_label(r"$ VI_{P_{XY}}$", fontsize=fonts)
        tax.right_axis_label(r"$ MI_{P_{XY}}  $", fontsize=fonts)
        tax.bottom_axis_label(r"$\Delta H_{P_{X}\cdot{P_{Y}}}$", fontsize=fonts)
    
    elif(all(hasCmetEntropicCoords(edf[l]) for l in range(len(edf)))):

        figure, tax = ternary_axes_subplot.figure(scale=scale)
        figure.set_size_inches(pltscale, pltscale)
        tax.boundary(linewidth=2.0)
        tax.gridlines(multiple = gridl, color="blue")

        marker = list(['X','v','o'])
        colors = get_cmap(len(edf))

        for i in range(3): 
            for j in range(len(edf)):
                points = entcoords(edf[j].take([i]), scale)
                mk = marker[i]*len(edf) ; names = str(edf[i].index[i]+' with '+str(len(edf)-j)+' PC')

                if(i != 1):
                    tax.scatter(points, s = s_mk , marker = mk[j], color=colors(j), linewidth='0.1')
                
                else:
                    tax.scatter(points, s = s_mk , marker = mk[j], color ='white' , edgecolors = colors(j), linewidth='1.5')
                
        if(chart_title == ""):
            tax.set_title("Channel Multivariate entropies (CMET)", fontsize = fonts + 5)
        else:
            tax.set_title(chart_title, fontsize = fonts + 5)

        tax.left_axis_label(r"$ VI'_{P_\overline{XY}}$", fontsize=fonts)
        tax.right_axis_label(r"$ 2\cdot{I'_{P_\overline{XY}}}  $", fontsize=fonts)
        tax.bottom_axis_label(r"$\Delta H'_{P_\overline{XY}}$", fontsize=fonts)
       
        leg = list()
        leg.append(mpatches.Patch(color='white', label='Feature Type'))
        
        for i in range(len(edf)):leg.append(mpatches.Patch(color=colors(i), label=str(len(edf)-i)+' Principal Components'))
        leg.append(mpatches.Patch(color='white', label=''))
        leg.append(mpatches.Patch(color='white', label='Entropy types'))
    
        leg.append(mlines.Line2D([], [], color='grey', marker='X', linestyle='None',markersize=15, label='X')); leg.append(mlines.Line2D([], [], color='grey', marker='v', linestyle='None',markersize=15, label='Y')); leg.append(mlines.Line2D([], [], color='grey', marker='o', linestyle='None',markersize=15, label='XY'))


    tax.ticks(axis='lbr', linewidth=1, multiple = gridl , fontsize = ticks_size )
    tax.clear_matplotlib_ticks()
    
    if(all(hasCmetEntropicCoords(edf[l]) for l in range(len(edf)))):
        
        tax.legend(handles=leg, loc = 1 ,labelspacing = 1 , fontsize = 10)
    else:
        tax.legend(title = 'Features' ,labelspacing = 1.5 , fontsize = 12)
    
    
    tax.show()
    figure.savefig('plot.pdf')
    #return tax
    
    
    
    
    
    


"""
Various Heatmaps.
"""

import functools
import numpy
from matplotlib import pyplot

from .helpers import unzip, normalize, simplex_iterator, project_point
from .colormapping import get_cmap, colormapper, colorbar_hack

### Heatmap Triangulation Coordinates

## Triangular Heatmaps


def blend_value(data, i, j, k, keys=None):
    """Computes the average value of the three vertices of a triangle in the
    simplex triangulation, where two of the vertices are on the lower
    horizontal."""

    key_size = len(list(data.keys())[0])
    if not keys:
        keys = triangle_coordinates(i, j, k)
    # Reduce key from (i, j, k) to (i, j) if necessary
    keys = [tuple(key[:key_size]) for key in keys]

    # Sum over the values of the points to blend
    try:
        s = sum(data[key] for key in keys)
        value = s / 3.
    except KeyError:
        value = None
    return value


def alt_blend_value(data, i, j, k):
    """Computes the average value of the three vertices of a triangle in the
    simplex triangulation, where two of the vertices are on the upper
    horizontal."""

    keys = alt_triangle_coordinates(i, j, k)
    return blend_value(data, i, j, k, keys=keys)


def triangle_coordinates(i, j, k):
    """
    Computes coordinates of the constituent triangles of a triangulation for the
    simplex. These triangules are parallel to the lower axis on the lower side.

    Parameters
    ----------
    i,j,k: enumeration of the desired triangle

    Returns
    -------
    A numpy array of coordinates of the hexagon (unprojected)
    """

    return [(i, j, k), (i + 1, j, k - 1), (i, j + 1, k - 1)]


def alt_triangle_coordinates(i, j, k):
    """
    Computes coordinates of the constituent triangles of a triangulation for the
    simplex. These triangules are parallel to the lower axis on the upper side.

    Parameters
    ----------
    i,j,k: enumeration of the desired triangle

    Returns
    -------
    A numpy array of coordinates of the hexagon (unprojected)
    """

    return [(i, j + 1, k - 1), (i + 1, j, k - 1), (i + 1, j + 1, k - 2)]


## Hexagonal Heatmaps ##

def generate_hexagon_deltas():
    """
    Generates a dictionary of the necessary additive vectors to generate the
    heaxagon points for the haxagonal heatmap.
    """

    zero = numpy.array([0, 0, 0])
    alpha = numpy.array([-1./3, 2./3, 0])
    deltaup = numpy.array([1./3, 1./3, 0])
    deltadown = numpy.array([2./3, -1./3, 0])
    i_vec = numpy.array([0, 1./2, -1./2])
    i_vec_down = numpy.array([1./2, -1./2, 0])
    deltaX_vec = numpy.array([1./2, 0, -1./2])

    d = dict()
    # Corner Points
    d["100"] = [zero, -deltaX_vec, -deltadown, -i_vec_down]
    d["010"] = [zero, i_vec_down, -alpha, -i_vec]
    d["001"] = [zero, i_vec, deltaup, deltaX_vec]
    # On the Edges
    d["011"] = [i_vec, deltaup, deltadown, -alpha, -i_vec]
    d["101"] = [-deltaX_vec, -deltadown, alpha, deltaup, deltaX_vec]
    d["110"] = [i_vec_down, -alpha, -deltaup, -deltadown, -i_vec_down]
    # Interior point
    d["111"] = [alpha, deltaup, deltadown, -alpha, -deltaup, -deltadown]

    return d


hexagon_deltas = generate_hexagon_deltas()


def hexagon_coordinates(i, j, k):
    """
    Computes coordinates of the constituent hexagons of a heaxagonal heatmap.

    Parameters
    ----------
    i, j, k: enumeration of the desired hexagon

    Returns
    -------
    A numpy array of coordinates of the hexagon (unprojected)
    """

    signature = ""
    for x in [i, j, k]:
        if x == 0:
            signature += "0"
        else:
            signature += "1"
    deltas = hexagon_deltas[signature]
    center = numpy.array([i, j, k])
    return numpy.array([center + x for x in deltas])


## Heatmaps ##

def polygon_generator(data, scale, style, permutation=None):
    """Generator for the vertices of the polygon to be colored and its color,
    depending on style. Called by heatmap."""

    # We'll project the coordinates inside this function to prevent
    # passing around permutation more than necessary
    project = functools.partial(project_point, permutation=permutation)

    if isinstance(data, dict):
        data_gen = data.items()
    else:
        # Only works with style == 'h'
        data_gen = data

    for key, value in data_gen:
        if value is None:
            continue
        i = key[0]
        j = key[1]
        k = scale - i - j
        if style == 'h':
            vertices = hexagon_coordinates(i, j, k)
            yield (map(project, vertices), value)
        elif style == 'd':
            # Upright triangles
            if (i <= scale) and (j <= scale) and (k >= 0):
                vertices = triangle_coordinates(i, j, k)
                yield (map(project, vertices), value)
            # Upside-down triangles
            if (i < scale) and (j < scale) and (k >= 1):
                vertices = alt_triangle_coordinates(i, j, k)
                value = blend_value(data, i, j, k)
                yield (map(project, vertices), value)
        elif style == 't':
            # Upright triangles
            if (i < scale) and (j < scale) and (k > 0):
                vertices = triangle_coordinates(i, j, k)
                value = blend_value(data, i, j, k)
                yield (map(project, vertices), value)
            # If not on the boundary add the upside-down triangle
            if (i < scale) and (j < scale) and (k > 1):
                vertices = alt_triangle_coordinates(i, j, k)
                value = alt_blend_value(data, i, j, k)
                yield (map(project, vertices), value)


def heatmap(data, scale, vmin=None, vmax=None, cmap=None, ax=None,
            scientific=False, style='triangular', colorbar=True,
            permutation=None, use_rgba=False, cbarlabel=None, cb_kwargs=None):
    """
    Plots heatmap of given color values.

    Parameters
    ----------
    data: dictionary
        A dictionary mapping the i, j polygon to the heatmap color, where
        i + j + k = scale.
    scale: Integer
        The scale used to partition the simplex.
    vmin: float, None
        The minimum color value, used to normalize colors. Computed if absent.
    vmax: float, None
        The maximum color value, used to normalize colors. Computed if absent.
    cmap: String or matplotlib.colors.Colormap, None
        The name of the Matplotlib colormap to use.
    ax: Matplotlib AxesSubplot, None
        The subplot to draw on.
    scientific: Bool, False
        Whether to use scientific notation for colorbar numbers.
    style: String, "triangular"
        The style of the heatmap, "triangular", "dual-triangular" or "hexagonal"
    colorbar: bool, True
        Show colorbar.
    permutation: string, None
        A permutation of the coordinates
    use_rgba: bool, False
        Use rgba color values
    cbarlabel: string, None
        Text label for the colorbar
    cb_kwargs: dict
        dict of kwargs to pass to colorbar

    Returns
    -------
    ax: The matplotlib axis
    """

    if not ax:
        fig, ax = pyplot.subplots()
    # If use_rgba, make the RGBA values numpy arrays so that they can
    # be averaged.
    if use_rgba:
        for k, v in data.items():
            data[k] = numpy.array(v)
    else:
        cmap = get_cmap(cmap)
        if vmin is None:
            vmin = min(data.values())
        if vmax is None:
            vmax = max(data.values())
    style = style.lower()[0]
    if style not in ["t", "h", 'd']:
        raise ValueError("Heatmap style must be 'triangular', 'dual-triangular', or 'hexagonal'")

    vertices_values = polygon_generator(data, scale, style,
                                        permutation=permutation)

    # Draw the polygons and color them
    for vertices, value in vertices_values:
        if value is None:
            continue
        if not use_rgba:
            color = colormapper(value, vmin, vmax, cmap=cmap)
        else:
            color = value  # rgba tuple (r,g,b,a) all in [0,1]
        # Matplotlib wants a list of xs and a list of ys
        xs, ys = unzip(vertices)
        ax.fill(xs, ys, facecolor=color, edgecolor=color)

    if not cb_kwargs:
        cb_kwargs = dict()
    if colorbar:
        colorbar_hack(ax, vmin, vmax, cmap, scientific=scientific,
                      cbarlabel=cbarlabel, **cb_kwargs)
    return ax


## User Convenience Functions ##


def heatmapf(func, scale=10, boundary=True, cmap=None, ax=None,
             scientific=False, style='triangular', colorbar=True,
             permutation=None, vmin=None, vmax=None, cbarlabel=None,
             cb_kwargs=None):
    """
    Computes func on heatmap partition coordinates and plots heatmap. In other
    words, computes the function on lattice points of the simplex (normalized
    points) and creates a heatmap from the values.

    Parameters
    ----------
    func: Function
        A function of 3-tuples to be heatmapped
    scale: Integer
        The scale used to partition the simplex
    boundary: Bool, True
        Include the boundary points or not
    cmap: String, None
        The name of the Matplotlib colormap to use
    ax: Matplotlib axis object, None
        The axis to draw the colormap on
    style: String, "triangular"
        The style of the heatmap, "triangular", "dual-triangular" or "hexagonal"
    scientific: Bool, False
        Whether to use scientific notation for colorbar numbers.
    colorbar: bool, True
        Show colorbar.
    permutation: string, None
        A permutation of the coordinates
    vmin: float
        The minimum color value, used to normalize colors.
    vmax: float
        The maximum color value, used to normalize colors.
    cb_kwargs: dict
        dict of kwargs to pass to colorbar

    Returns
    -------
    ax, The matplotlib axis
    """

    # Apply the function to a simplex partition
    data = dict()
    for i, j, k in simplex_iterator(scale=scale, boundary=boundary):
        data[(i, j)] = func(normalize([i, j, k]))
    # Pass everything to the heatmapper
    ax = heatmap(data, scale, cmap=cmap, ax=ax, style=style,
                 scientific=scientific, colorbar=colorbar,
                 permutation=permutation, vmin=vmin, vmax=vmax, 
                 cbarlabel=cbarlabel, cb_kwargs=cb_kwargs)
    return ax


def svg_polygon(coordinates, color):
    """
    Create an svg triangle for the stationary heatmap.

    Parameters
    ----------
    coordinates: list
        The coordinates defining the polygon
    color: string
        RGB color value e.g. #26ffd1

    Returns
    -------
    string, the svg string for the polygon
    """

    coord_str = []
    for c in coordinates:
        coord_str.append(",".join(map(str, c)))
    coord_str = " ".join(coord_str)
    polygon = '<polygon points="%s" style="fill:%s;stroke:%s;stroke-width:0"/>\n' % (coord_str, color, color)
    return polygon


def svg_heatmap(data, scale, filename, vmax=None, vmin=None, style='h',
                permutation=None, cmap=None):
    """
    Create a heatmap in SVG format. Intended for use with very large datasets,
    which would require large amounts of RAM using matplotlib. You can convert
    the image to another format with e.g. ImageMagick:

    convert -density 1200 -resize -rotate 180 1000x1000 your.svg your.png

    Parameters
    ----------

    data: dictionary or k, v generator
        A dictionary mapping the i, j polygon to the heatmap color, where
        i + j + k = scale. If using a generator, style must be 'h'.
    scale: Integer
        The scale used to partition the simplex.
    filename: string
        The filename to write the SVG data to.
    vmin: float
        The minimum color value, used to normalize colors.
    vmax: float
        The maximum color value, used to normalize colors.
    cmap: String or matplotlib.colors.Colormap, None
        The name of the Matplotlib colormap to use.
    style: String, "h"
        The style of the heatmap, "triangular", "dual-triangular" or "hexagonal"
    permutation: string, None
        A permutation of the coordinates
    """

    style = style.lower()[0]
    if style not in ["t", "h", 'd']:
        raise ValueError("Heatmap style must be 'triangular', 'dual-triangular', or 'hexagonal'")

    if not isinstance(data, dict):
        if not style == 'h':
            raise ValueError("Data can only be given as a generator for hexagonal style heatmaps because of blending for adjacent polygons.")
        elif vmax is None or vmin is None:
            raise ValueError("vmax and vmin must be supplied for data given as a generator.")

    cmap = get_cmap(cmap)

    if not vmin:
        vmin = min(data.values())
    if not vmax:
        vmax = max(data.values())

    height = scale * numpy.sqrt(3) / 2 + 2

    output_file = open(filename, 'w')
    output_file.write('<svg height="%s" width="%s">\n' % (height, scale))

    vertices_values = polygon_generator(data, scale, style,
                                        permutation=permutation)

    # Draw the polygons and color them
    for vertices, value in vertices_values:
        color = colormapper(value, vmin, vmax, cmap=cmap)
        output_file.write(svg_polygon(vertices, color))

    output_file.write('</svg>\n')



"""
Plotting functions: scatter, plot (curves), axis labelling.
"""

import matplotlib
from matplotlib import pyplot
import numpy as np

from .helpers import project_sequence
from .colormapping import get_cmap, colorbar_hack


### Drawing Helpers ###

def resize_drawing_canvas(ax, scale=1., angle = 0.):
    """
    Makes sure the drawing surface is large enough to display projected
    content.

    Parameters
    ----------
    ax: Matplotlib AxesSubplot, None
        The subplot to draw on.
    scale: float, 1.0
        Simplex scale size.
    """ 
    
    if(angle == 0):
        ax.set_ylim((-0.10 * scale, .90 * scale))
        ax.set_xlim((-0.05 * scale, 1.05 * scale))
        
    else:
        ax.set_ylim((-0.95 * scale, .20 * scale))
        ax.set_xlim((-0.1 * scale, 1.1 * scale))

def clear_matplotlib_ticks(ax=None, axis="both"):
    """
    Clears the default matplotlib axes, or the one specified by the axis
    argument.

    Parameters
    ----------
    ax: Matplotlib AxesSubplot, None
        The subplot to draw on.
    axis: string, "both"
        The axis to clear: "x" or "horizontal", "y" or "vertical", or "both"
    """
    if not ax:
        return
    if axis.lower() in ["both", "x", "horizontal"]:
        ax.set_xticks([], [])
    if axis.lower() in ["both", "y", "vertical"]:
        ax.set_yticks([], [])


## Curve Plotting ##

def plot(points, ax=None, permutation=None, angle = None, **kwargs):
    """
    Analogous to maplotlib.plot. Plots trajectory points where each point is a
    tuple (x,y,z) satisfying x + y + z = scale (not checked). The tuples are
    projected and plotted as a curve.

    Parameters
    ----------
    points: List of 3-tuples
        The list of tuples to be plotted as a connected curve.
    ax: Matplotlib AxesSubplot, None
        The subplot to draw on.
    kwargs:
        Any kwargs to pass through to matplotlib.
    """
    if not ax:
        fig, ax = pyplot.subplots()
    xs, ys = project_sequence(points, permutation=permutation, angle = angle)
    ax.plot(xs, ys, **kwargs)
    return ax


def plot_colored_trajectory(points, cmap=None, ax=None, permutation=None, angle = None,
                            **kwargs):
    """
    Plots trajectories with changing color, simlar to `plot`. Trajectory points
    are tuples (x,y,z) satisfying x + y + z = scale (not checked). The tuples are
    projected and plotted as a curve.

    Parameters
    ----------
    points: List of 3-tuples
        The list of tuples to be plotted as a connected curve.
    ax: Matplotlib AxesSubplot, None
        The subplot to draw on.
    cmap: String or matplotlib.colors.Colormap, None
        The name of the Matplotlib colormap to use.
    kwargs:
        Any kwargs to pass through to matplotlib.
    """
    if not ax:
        fig, ax = pyplot.subplots()
    cmap = get_cmap(cmap)
    xs, ys = project_sequence(points, permutation=permutation, angle = angle)

    # We want to color each segment independently...which is annoying.
    segments = []
    for i in range(len(xs) - 1):
        cur_line = []
        x_before = xs[i]
        y_before = ys[i]
        x_after = xs[i+1]
        y_after = ys[i+1]

        cur_line.append([x_before, y_before])
        cur_line.append([x_after, y_after])
        segments.append(cur_line)
    segments = np.array(segments)

    line_segments = matplotlib.collections.LineCollection(segments, cmap=cmap, **kwargs)
    line_segments.set_array(np.arange(len(segments)))
    ax.add_collection(line_segments)

    return ax


def scatter(points, ax=None, permutation=None, colorbar=False, colormap=None,
            vmin=0, vmax=1, scientific=False, cbarlabel=None, cb_kwargs=None,
            angle = None, **kwargs):
    """
    Plots trajectory points where each point satisfies x + y + z = scale.
    First argument is a list or numpy array of tuples of length 3.

    Parameters
    ----------
    points: List of 3-tuples
        The list of tuples to be scatter-plotted.
    ax: Matplotlib AxesSubplot, None
        The subplot to draw on.
    colorbar: bool, False
        Show colorbar.
    colormap: String or matplotlib.colors.Colormap, None
        The name of the Matplotlib colormap to use.
    vmin: int, 0
        Minimum value for colorbar.
    vmax: int, 1
        Maximum value for colorbar.
    cb_kwargs: dict
        Any additional kwargs to pass to colorbar
    kwargs:
        Any kwargs to pass through to matplotlib.
    """
    if not ax:
        fig, ax = pyplot.subplots()
    xs, ys = project_sequence(points, permutation=permutation, angle = angle)
    ax.scatter(xs, ys, vmin=vmin, vmax=vmax, **kwargs)

    if colorbar and (colormap != None):
        if cb_kwargs != None:
            colorbar_hack(ax, vmin, vmax, colormap, scientific=scientific,
                          cbarlabel=cbarlabel, **cb_kwargs)
        else:
            colorbar_hack(ax, vmin, vmax, colormap, scientific=scientific,
                          cbarlabel=cbarlabel)

    return ax



"""
Wrapper class for all ternary plotting functions.
"""

from functools import partial

import numpy
from matplotlib import pyplot

from . import heatmapping
from . import lines
from . import plotting
from .helpers import project_point, convert_coordinates_sequence


def figure(ax=None, scale=None, permutation=None, angle = None):
    """
    Wraps a Matplotlib AxesSubplot or generates a new one. Emulates matplotlib's
    > figure, ax = pyplot.subplots()

    Parameters
    ----------
    ax: AxesSubplot, None
        The matplotlib AxesSubplot to wrap
    scale: float, None
        The scale factor of the ternary plot
    """
    ternary_ax = TernaryAxesSubplot(ax=ax, scale=scale, permutation=permutation, angle=angle)
    return ternary_ax.get_figure(), ternary_ax


def mpl_redraw_callback(event, tax):
    """
    Callback to properly rotate and redraw text labels when the plot is drawn
    or resized.

    Parameters
    ----------
    event: a matplotlib event
        either 'resize_event' or 'draw_event'
    tax: TernaryAxesSubplot
         the TernaryAxesSubplot
    """
    tax._redraw_labels()


class TernaryAxesSubplot(object):
    """
    Wrapper for python-ternary and matplotlib figure. Parameters for member
    functions simply pass through to ternary's functions with the same names.
    This class manages the matplotlib axes, the scale, and the boundary scale
    to ease the use of ternary plotting functions.
    """

    def __init__(self, ax=None, scale=None, permutation=None , angle=None):
        
        if not scale:
            scale = 1.0

        if not angle:
            angle = 0.0

        if ax:
            self.ax = ax
        else:
            _, self.ax = pyplot.subplots()

        self.set_scale(scale=scale)
        self.set_angle(angle = angle)
        self._permutation = permutation
        self._boundary_scale = scale
        self._boundary_angle = angle
        # Container for the axis labels supplied by the user
        self._labels = dict()
        self._corner_labels = dict()
        self._ticks = dict()
        # Container for the redrawing of labels
        self._to_remove = []
        self._connect_callbacks()

    def _connect_callbacks(self):
        """Connect resize matplotlib callbacks."""
        figure = self.get_figure()
        callback = partial(mpl_redraw_callback, tax=self)
        event_names = ('resize_event', 'draw_event')
        for event_name in event_names:
            figure.canvas.mpl_connect(event_name, callback)

    def __repr__(self):
        return "TernaryAxesSubplot: %s" % self.ax.__hash__()

    def get_axes(self):
        """Returns the underlying matplotlib AxesSubplot object."""
        return self.ax

    def get_figure(self):
        """Return the underlying matplotlib figure object."""
        ax = self.get_axes()
        return ax.get_figure()

    def set_scale(self, scale=None):
        self._scale = scale
        self.resize_drawing_canvas()

    def get_scale(self):
        return self._scale

    def set_angle(self, angle=None):
        self._angle = angle

    def get_angle(self):
        return self._angle


    def set_axis_limits(self, axis_limits=None):
        """
        Set min and max data limits for each of the three axes.

        axis_limits = dict
            keys are 'b','l' and 'r' for the three axes
            vals are lists of the min and max values for the axis in
            data units.
        """
        self._axis_limits = axis_limits

    def get_axis_limits(self):
        return self._axis_limits

    # Title and Axis Labels

    def set_title(self, title, **kwargs):
        """Sets the title on the underlying matplotlib AxesSubplot."""
        ax = self.get_axes()
        ax.set_title(title, **kwargs)

    def left_axis_label(self, label, position=None,  rotation=60, offset=0.08,
                        **kwargs):
        """
        Sets the label on the left axis.

        Parameters
        ----------
        label: String
            The axis label
        position: 3-Tuple of floats, None
            The position of the text label
        rotation: float, 60
            The angle of rotation of the label
        offset: float,
            Used to compute the distance of the label from the axis
        kwargs:
            Any kwargs to pass through to matplotlib.
        """
        angle = self.get_angle()

        if not position:
            position = (-offset, 3./5, 2./5)
        
        if (angle != 0):
            rotation = -60
            position = (-.45*offset, 2.5/5,0)

        self._labels["left"] = (label, position, rotation, kwargs)

    def right_axis_label(self, label, position=None, rotation=-60, offset=0.08,
                         **kwargs):


        """
        Sets the label on the right axis.

        Parameters
        ----------
        label: String
            The axis label
        position: 3-Tuple of floats, None
            The position of the text label
        rotation: float, -60
            The angle of rotation of the label
        offset: float,
            Used to compute the distance of the label from the axis
        kwargs:
            Any kwargs to pass through to matplotlib.
        """
        angle = self.get_angle()
        
        if not position:
            position = (2. / 5 + offset, 3. / 5, 0)
            
        if (angle != 0):
            rotation = 60
            position =  (2./5 + 1.53*offset,  2.5/ 5 ,0)   
        
        self._labels["right"] = (label, position, rotation, kwargs)

    def bottom_axis_label(self, label, position=None, rotation=0, offset=0.02,
                          **kwargs):
        """
        Sets the label on the bottom axis.

        Parameters
        ----------
        label: String
            The axis label
        position: 3-Tuple of floats, None
            The position of the text label
        rotation: float, 0
            The angle of rotation of the label
        offset: float,
            Used to compute the distance of the label from the axis
        kwargs:
            Any kwargs to pass through to matplotlib.
        """
        angle = self.get_angle()
        
        if not position:
            position = (0.5-1.3*offset, 2* offset, 0.5)

        if (angle != 0):
            offset=0.08
            position =  (0.005, 0.843 + 2.15 * offset, 0)

        self._corner_labels["bottom"] = (label, position, rotation, kwargs)

    def right_corner_label(self, label, position=None, rotation=0, offset=0.08,
                           **kwargs):
        """
        Sets the label on the right corner (complements left axis).

        Parameters
        ----------
        label: String
            The axis label
        position: 3-Tuple of floats, None
            The position of the text label
        rotation: float, 0
            The angle of rotation of the label
        offset: float,
            Used to compute the distance of the label from the axis
        kwargs:
            Any kwargs to pass through to matplotlib.
        """

        if not position:
            position = (1, offset / 2, 0)
        self._labels["right"] = (label, position, rotation, kwargs)

    def left_corner_label(self, label, position=None, rotation=0, offset=0.08,
                          **kwargs):
        """
        Sets the label on the left corner (complements right axis.)

        Parameters
        ----------
        label: string
            The axis label
        position: 3-Tuple of floats, None
            The position of the text label
        rotation: float, 0
            The angle of rotation of the label
        offset: float,
            Used to compute the distance of the label from the axis
        kwargs:
            Any kwargs to pass through to matplotlib.
        """

        if not position:
            position = (-offset / 2, offset / 2, 0)
        self._corner_labels["left"] = (label, position, rotation, kwargs)

    def top_corner_label(self, label, position=None, rotation=0, offset=0.2,
                         **kwargs):
        """
        Sets the label on the bottom axis.

        Parameters
        ----------
        label: String
            The axis label
        position: 3-Tuple of floats, None
            The position of the text label
        rotation: float, 0
            The angle of rotation of the label
        offset: float,
            Used to compute the distance of the label from the axis
        kwargs:
            Any kwargs to pass through to matplotlib.
        """

        if not position:
            position = (-offset / 2, 1 + offset, 0)
        self._corner_labels["bottom"] = (label, position, rotation, kwargs)

    def annotate(self, text, position, **kwargs):
        ax = self.get_axes()
        p = project_point(position)
        ax.annotate(text, (p[0], p[1]), **kwargs)

    # Boundary and Gridlines

    def boundary(self, scale=None, axes_colors=None, angle = None, **kwargs):
        # Sometimes you want to draw a bigger boundary
        if not scale:
            scale = self._boundary_scale  # defaults to self._scale
        if not angle:
            angle = self._boundary_angle
        ax = self.get_axes()
        self.resize_drawing_canvas(scale,angle)
        lines.boundary(scale=scale, ax=ax, axes_colors=axes_colors,angle = angle, **kwargs)

    def gridlines(self, multiple=None, horizontal_kwargs=None, left_kwargs=None,
                  right_kwargs=None, **kwargs):
        ax = self.get_axes()
        scale = self.get_scale()
        angle = self.get_angle()
        lines.gridlines(scale=scale, multiple=multiple,
                        ax=ax, horizontal_kwargs=horizontal_kwargs,
                        left_kwargs=left_kwargs, right_kwargs=right_kwargs, angle = angle, 
                        **kwargs)

    # Various Lines

    def line(self, p1, p2, **kwargs):
        ax = self.get_axes()
        angle = self.get_angle()
        lines.line(ax, p1, p2, angle, **kwargs)

    def horizontal_line(self, i, **kwargs):
        ax = self.get_axes()
        scale = self.get_scale()
        angle = self.get_angle()
        lines.horizontal_line(ax, scale, i, angle, **kwargs)

    def left_parallel_line(self, i, **kwargs):
        ax = self.get_axes()
        scale = self.get_scale()
        angle = self.get_angle()
        lines.left_parallel_line(ax, scale, i, angle, **kwargs)

    def right_parallel_line(self, i, **kwargs):
        ax = self.get_axes()
        scale = self.get_scale()
        angle = self.get_angle()
        lines.right_parallel_line(ax, scale, i, angle, **kwargs)

    # Matplotlib passthroughs

    def close(self):
        fig = self.get_figure()
        pyplot.close(fig)

    def legend(self, *args, **kwargs):
        ax = self.get_axes()
        ax.legend(*args, **kwargs)

    def savefig(self, filename, dpi=200, format=None):
        self._redraw_labels()
        fig = self.get_figure()
        fig.savefig(filename, format=format, dpi=dpi)

    def show(self):
        self._redraw_labels()
        pyplot.show()

    # Axis ticks

    def clear_matplotlib_ticks(self, axis="both"):
        """Clears the default matplotlib ticks."""
        ax = self.get_axes()
        plotting.clear_matplotlib_ticks(ax=ax, axis=axis)

    def get_ticks_from_axis_limits(self, multiple=1):
        """
        Taking self._axis_limits and self._boundary_scale get the scaled
        ticks for all three axes and store them in self._ticks under the
        keys 'b' for bottom, 'l' for left and 'r' for right axes.
        """
        for k in ['b', 'l', 'r']:
            self._ticks[k] = numpy.linspace(
                self._axis_limits[k][0],
                self._axis_limits[k][1],
                self._boundary_scale / float(multiple) + 1
            ).tolist()

    def set_custom_ticks(self, locations=None, clockwise=False, multiple=1,
                         axes_colors=None, tick_formats=None, **kwargs):
        """
        Having called get_ticks_from_axis_limits, set the custom ticks on the
        plot.
        """
        for k in ['b', 'l', 'r']:
            self.ticks(ticks=self._ticks[k], locations=locations,
                       axis=k, clockwise=clockwise, multiple=multiple,
                       axes_colors=axes_colors, tick_formats=tick_formats,
                       **kwargs)

    def ticks(self, ticks=None, locations=None, multiple=1, axis='blr',
              clockwise=False, axes_colors=None, tick_formats=None, **kwargs):
        ax = self.get_axes()
        scale = self.get_scale()
        angle = self.get_angle()
        lines.ticks(ax, scale, ticks=ticks, locations=locations,
                    multiple=multiple, clockwise=clockwise, axis=axis,
                    axes_colors=axes_colors, tick_formats=tick_formats, angle = angle,
                    **kwargs)

    # Redrawing and resizing

    def resize_drawing_canvas(self, scale=None, angle = None):
        ax = self.get_axes()
        if not scale:
            scale = self.get_scale()
        plotting.resize_drawing_canvas(ax, scale=scale, angle = angle)

    def _redraw_labels(self):
        """Redraw axis labels, typically after draw or resize events."""
        ax = self.get_axes()
        # Remove any previous labels
        for mpl_object in self._to_remove:
            mpl_object.remove()
        self._to_remove = []
        # Redraw the labels with the appropriate angles
        label_data = list(self._labels.values())
        label_data.extend(self._corner_labels.values())
        for (label, position, rotation, kwargs) in label_data:
            transform = ax.transAxes
            x, y = project_point(position)
            # Calculate the new angle.
            position = numpy.array([x, y])
            new_rotation = ax.transData.transform_angles(
                numpy.array((rotation,)), position.reshape((1, 2)))[0]
            text = ax.text(x, y, label, rotation=new_rotation,
                           transform=transform, horizontalalignment="center",
                           **kwargs)
            text.set_rotation_mode("anchor")
            self._to_remove.append(text)

    def convert_coordinates(self, points, axisorder='blr'):
        """
        Convert data coordinates to simplex coordinates for plotting
        in the case that axis limits have been applied.
        """
        return convert_coordinates_sequence(points,self._boundary_scale,
                                            self._axis_limits, axisorder)

    # Various Plots

    def scatter(self, points, **kwargs):
        ax = self.get_axes()
        angle = self.get_angle()
        permutation = self._permutation
        plot_ = plotting.scatter(points, ax=ax, permutation=permutation, angle = angle,
                                 **kwargs)
        return plot_

    def plot(self, points, **kwargs):
        ax = self.get_axes()
        angle = self.get_angle()
        permutation = self._permutation
        plotting.plot(points, ax=ax, permutation=permutation, angle = angle,
                      **kwargs)

    def plot_colored_trajectory(self, points, cmap=None, **kwargs):
        ax = self.get_axes()
        angle = self.get_angle()
        permutation = self._permutation
        plotting.plot_colored_trajectory(points, cmap=cmap, ax=ax,
                                         permutation=permutation, angle=angle, **kwargs)

    def heatmap(self, data, scale=None, cmap=None, scientific=False,
                style='triangular', colorbar=True, use_rgba=False,
                vmin=None, vmax=None, cbarlabel=None, cb_kwargs=None):
        permutation = self._permutation
        if not scale:
            scale = self.get_scale()
        if style.lower()[0] == 'd':
            self._boundary_scale = scale + 1
        ax = self.get_axes()
        heatmapping.heatmap(data, scale, cmap=cmap, style=style, ax=ax,
                            scientific=scientific, colorbar=colorbar,
                            permutation=permutation, use_rgba=use_rgba,
                            vmin=vmin, vmax=vmax, cbarlabel=cbarlabel,
                            cb_kwargs=cb_kwargs)

    def heatmapf(self, func, scale=None, cmap=None, boundary=True,
                 style='triangular', colorbar=True, scientific=True,
                 vmin=None, vmax=None, cbarlabel=None, cb_kwargs=None):
        if not scale:
            scale = self.get_scale()
        if style.lower()[0] == 'd':
            self._boundary_scale = scale + 1
        permutation = self._permutation
        ax = self.get_axes()
        heatmapping.heatmapf(func, scale, cmap=cmap, style=style,
                             boundary=boundary, ax=ax, scientific=scientific,
                             colorbar=colorbar, permutation=permutation,
                             vmin=vmin, vmax=vmax, cbarlabel=cbarlabel,
                             cb_kwargs=cb_kwargs)



'''

Auxiliar functions used in the calculation of the Measures

'''

import numpy as np                  
import pandas as pd
import operator

from sys import exit as exit
from warnings import warn as warning
from scipy.stats import entropy as entropy
from sklearn.preprocessing import LabelEncoder   #' Used for discretization
from functools import reduce


def discretization (df , nbins = 1):

    """
    Function created for the discretization of a dataframe in #nbins equalwidth bins

    > comprobation = discretization(df,nbins)

    Parameters
    ----------
    df : raw DataFrame 
    nbins : number of bins 

    Returns
    ----------
    df : Discretized Dataframe

    """


    if(not isinstance(df,pd.DataFrame)):
        exit("Can only work with Data Frames!")
    #' Rows or columns equal to zero?
    if (len(df.columns) == 0 or len(df.index) == 0): 
        exit("Can only work with non-empty Data Frames!")
    
    if(nbins == 1):
        nbins = int(len(df.index)**(1/3))

    if (not isinstance(nbins,np.int)):
        nbins = int(nbins)
    
    #Object -> String 
    #Numeric -> (Float o int)
    warning("Discretizing data!")
    lb = LabelEncoder()
    
    df.update(df.select_dtypes(exclude = ['float64',np.int]).apply((lambda x : lb.fit_transform(x)), axis = 0))
    df.update(df.select_dtypes(include = ['float64',np.int]).apply((lambda x : pd.cut(x, bins = nbins , labels = list(range(nbins))))).astype('object'))

    disc = df.astype('category')

    return disc


def sjoin(df,lis,sep=''):

    """
    Function created for merge all the colums of a DataFrame in one column

    > odf = sjoin(df,df.columns,'')

    Parameters
    ----------
    df : Raw DataFrame 
    lis: Columns to merge
    sep : Separator
    Returns
    ----------
    odf : Merged DataFrame

    """

    return reduce(lambda x, y: x.astype(str).str.cat(y.astype(str), sep=sep),[df[col] for col in lis])


def ent(data , base = 2):
    
    """
    Function created for calculating the entropy of a pandas Series

    > entropy = ent(df,base = 2)

    Parameters
    ----------
    df : Raw DataFrame 
    base : Logarithm base used for the entropy calculation (Default value equal 2)
    ----------
    entropy : Value (float)

    """
   
    if(isinstance(data, pd.DataFrame)):
        data = sjoin(data,lis = data.columns)

    p_data = data.value_counts()/len(data) # calculates the probabilities
    entr = np.nan_to_num(entropy(p_data , base = base))  # input probabilities to get the entropy 
    
    return entr



def condentropy(X,Y = None, base = 2):

    """
    Function created for calculating the conditional entropy of two dataframes
    Formula used from the R function condentropy
    https://rdrr.io/cran/infotheo/man/condentropy.html
    
    H(X|Y) = H(Y,X)-H(Y)
    
    > entropy_value = condentropy(X,Y,base)

    Parameters
    ----------
    X : DataFrame 
    Y : DataFrame
    base : Logarithm base used for the entropy calculation (Default value equal 2)
    ----------
    entropy_value : Value (float)

    """
    
    if(isinstance(Y, type(None))):
        Hres = ent(X,base = base)
        
    else:
        
        if(isinstance(Y, pd.Series)):
            Y = Y.to_frame(name = None)
        if(isinstance(X, pd.Series)):
            X = X.to_frame(name = None)
        
        joint = pd.merge(Y,X, left_index = True, right_index = True)
        Hyx = ent(joint,base = base)
        Hy = ent(Y,base = base) 
        
        Hres = Hyx - Hy
        
    return Hres




from matplotlib import pyplot as plt

from .plotting import (
    clear_matplotlib_ticks,
    plot,
    resize_drawing_canvas,
    scatter,
)

from .lines import (
    boundary,
    gridlines,
    line,
    horizontal_line,
    left_parallel_line,
    right_parallel_line,
)

from .helpers import project_point

from .colormapping import get_cmap
from .heatmapping import heatmap, heatmapf, svg_heatmap

from .ternary_axes_subplot import figure, TernaryAxesSubplot


'''

Functions used for calculating entropic measures (CMET Triangle)

'''

from sys import exit as exit
from warnings import warn as warning
from scipy.stats import entropy as entropy

from .auxfunc import * 


def jentropies(X,Y , base = 2 , nbins = 1):

    """
    Channel Entropy decomposition of two dataframes X and Y

    > edf = jentropies_df (X,Y, base = 2)

    Parameters
    ----------
    X, Y : DataFrame for calculating the channel entropies
    base : The logarithm to be used in working out the sentropies (Default base = 2 -> "log2")

    Returns
    ----------
    edf : Pandas DataFrame containing the values of the entropies calculated from the contingency matrix

    """

    dimx = X.shape ; dimy = Y.shape
    
    if(not isinstance(X,pd.DataFrame)):
        exit("Can only work with Data Frames! (X it´s not a DataFrame)")
    
    if(not isinstance(Y,pd.DataFrame)):
        exit("Can only work with Data Frames! (Y it´s not a DataFrame)")   
    
    if (dimx[0] == 0 or dimx[1] == 0):
        exit("Can only work with non-empty data.frames X!")

    if (dimy[0] == 0 or dimy[1] == 0):
        exit("Can only condition on non-empty data.frame Y!")

    if (dimx[0] != dimy[0]):	
        exit("Can only condition on variable lists with the same number of instances!")

    if (not(all(X.dtypes)=='category')):
        warning("Discretizing data from X DataFrame before entropy calculation!") #' Throwing a Warning for communicating a discretization of data
        X = discretization(X , nbins)

    if (not(all(Y.dtypes)=='category')):
        warning("Discretizing data from X DataFrame before entropy calculation!") #' Throwing a Warning for communicating a discretization of data
        Y = discretization(Y , nbins)

    H_U = np.array([np.sum(X.apply((lambda x : np.log2(len(pd.unique(x)))), axis = 0).values) , np.sum(Y.apply((lambda x : np.log2(len(pd.unique(x)))), axis = 0).values)])
    H_P = np.array([ent(sjoin(X,lis = X.columns)),ent(sjoin(Y,lis = Y.columns))])
    
    #Calculation of Conditional Entropies of the Random Vectors
    VI_P = np.array([condentropy(X,Y,base = base),condentropy(Y,X,base = base)])
    
    edf = pd.DataFrame({'Type' : ['X','Y'], 'H_U': H_U , 'H_P': H_P , 'DeltaH_P' : H_U - H_P, }, columns = ['Type','H_U','H_P','DeltaH_P'])
    edf['M_P'] = H_P - VI_P
    edf['VI_P'] = VI_P
    edf = edf.set_index('Type')
    edf.loc['XY'] = edf.sum(axis = 0)

    return edf


def jentropies_binary(Nxy, base = 2):

    """
    Joint Entropy decomposition of a 2d contingency matrix due to a channel transmission

    > edf = jentropies_binary (Nxy, base = 2)

    Parameters
    ----------
    Nxy : Contingency matrix for calculating the channel entropies
    base : The logarithm to be used in working out the sentropies (Default base = 2 -> "log2")

    Returns
    ----------
    edf : Pandas DataFrame containing the values of the entropies calculated from the contingency matrix

    """

    dims = Nxy.shape # dim is a tuple wiht dim[0] rows  and dim[1] columns

    if(len(dims) != 2):
        exit("It must be a 2 dimensions array")

    Nx = Nxy.sum(axis = 1) ; Ny = Nxy.sum(axis = 0) 
    Hx = entropy(Nx , base = base) ; Hy = entropy(Ny , base = base)
    Ux = np.log2(dims[0]) ; Uy = np.log2(dims[1])
    Hxy = entropy(np.ravel(Nxy),base = base)
    VI_P = [(Hxy-Hy),(Hxy-Hx)]

    edf = pd.DataFrame({'Type':['X','Y'],'H_P2':[Hx,Hy] , 'H_U2' : [Ux,Uy] ,'DeltaH_P2':[(Ux - Hx) , (Uy - Hy)] , 'M_P2': [Hx - VI_P[0] , Hy - VI_P[1]] , 'VI_P2': VI_P }, columns = ['Type','H_U2','H_P2','DeltaH_P2','M_P2','VI_P2'])
    edf = edf.append(edf.apply(lambda x: np.sum(x,axis=0),axis = 0), ignore_index=True)
    #edf.loc['All'] = edf.sum(axis = 0)
    edf = edf.set_index('Type')

    return edf


'''

Functions used for calculating variables used in the Entropy Triangle Plotting phase

'''

from numpy import nan_to_num as nan_to_num
from warnings import warn as warning
import pandas as pd          # DataFrames manipulation
import matplotlib.pyplot as plt


#Definition of the variables


#SMET
derivedSplitSmetCoords = ["DeltaH_Pxi", "M_Pxi", "VI_Pxi"] # Multivariate entropic coordinates
aggregateSmetCoords = ["H_Ux", "DeltaH_Px", "M_Px", "VI_Px"] # Source multivariate aggregate coordinates (Sum of derivedSplitSmetCoords)
dualAggregateSmetCoords = ["H_Ux", "DeltaH_Px", "D_Px", "VI_Px"] # SMET coords without C_P_X

#CBET
cbetEntropicCoords = ["H_U2", "H_P2", "DeltaH_P2", "M_P2", "VI_P2"] # Caracterization of the variables in a DataFrame of Channel Multivariate Entropies

#CMET
cmetEntropicCoords = ["H_U", "H_P", "DeltaH_P", "M_P", "VI_P"] # Caracterization of the variables in a DataFrame of Channel Multivariate Entropies



#' Functions to detect SMET coordinates

def hasSplitSmetCoords(df):

    """
    A function to detect if the source multivariate split entropies are present: this enables working out the multivariate split entropic coordinates (SMET)
    derivedSplitSmetCoords = ["DeltaH_Pxi", "M_Pxi", "VI_Pxi", "Name"]

    > comprobation = hasSplitSmetCoords(df)

    Parameters
    ----------
    df : DataFrame with entropic variables

    Returns
    ----------
    Boolean: (True or False) In order to check if all the variables at the input dataframe correspond to this type of coordinates

    """
    return (df.columns.isin(derivedSplitSmetCoords).sum() == len(derivedSplitSmetCoords))



def hasAggregateSmetCoords(df):	

    """
    A function to detect if the source multivariate aggregate entropic coordinates are present (SMET)
    aggregateSmetCoords = ["H_Ux", "DeltaH_Px", "M_Px", "VI_Px"]

    > comprobation = hasAggregateSmetCoords(df)

    Parameters
    ----------
    df : DataFrame with entropic variables

    Returns
    ----------
    Boolean: (True or False) In order to check if all the variables at the input dataframe correspond to this type of coordinates

    """
    return (df.columns.isin(aggregateSmetCoords).sum() == len(aggregateSmetCoords))


def hasDualAggregateSmetCoords(df):

    """
    A function to detect if the source multivariate dual aggregate entropic coordinates are present (SMET)
    dualAggregateSmetCoords = ["H_Ux", "DeltaH_Px", "D_Px", "VI_Px"]

    > comprobation = hasDualAggregateSmetCoords(df)

    Parameters
    ----------
    df : DataFrame with entropic variables

    Returns
    ----------
    Boolean: (True or False) In order to check if all the variables at the input dataframe correspond to this type of coordinates

    """
    return (df.columns.isin(dualAggregateSmetCoords).sum() == len(dualAggregateSmetCoords))



#' Functions to detect CMET coordinates

def hasCbetEntropicCoords(df):

    """
    A function to detect if the the channel binary entropic coordinates are present (CBET)
    cmetEntropicCoords = ["H_U2", "H_P2", "DeltaH_P2", "M_P2", "VI_P2"]

    > comprobation = hasCbetEntropicCoords(df)

    Parameters
    ----------
    df : DataFrame with entropic variables

    Returns
    ----------
    Boolean: (True or False) In order to check if all the variables at the input dataframe correspond to this type of coordinates

    """

    #return all(df.columns.isin(cmetEntropicCoords))
    return (df.columns.isin(cbetEntropicCoords).sum() == len(cbetEntropicCoords))


#' Functions to detect CMET coordinates

def hasCmetEntropicCoords(df):

    """
    A function to detect if the the channel multivariate entropic coordinates are present (CMET)
    cmetEntropicCoords = ["H_U", "H_P", "DeltaH_P", "M_P", "VI_P"]

    > comprobation = hasCmetEntropicCoords(df)

    Parameters
    ----------
    df : DataFrame with entropic variables

    Returns
    ----------
    Boolean: (True or False) In order to check if all the variables at the input dataframe correspond to this type of coordinates

    """

    #return all(df.columns.isin(cmetEntropicCoords))
    return (df.columns.isin(cmetEntropicCoords).sum() == len(cmetEntropicCoords))



def entcoords(df,scale=100):

    """
    A function for calculating the normalized coordinates of the entropic measures for some DataFrame.
    It can be used for SMET and CMET cases, returning a list of arrays with the normalized scaled measures

    > entropies_coordinates = entcoords(df)
    >
    > Example:
    >
    > entropies_coordinates = [array([   9.24959601,  142.19609928,    2.96110915]), array([  15.4082864 ,  119.15900389,    1.60549331])]
    >

    Parameters
    ----------
    df : DataFrame with entropic variables
    scale : The scale used for plotting the triangle

    Returns
    ----------
    coor : List with the entropic measures to plot in the De Finneti Diagram

    """

    coor = list()

    for i in range(df.shape[0]):

        if(nan_to_num(df.iloc[i].values[2:6]).sum()>1):
            coor.append((df.iloc[i].values[2:6])/df.iloc[i].values[0])

        else:
            coor.append(df.iloc[i].values[2:6])

    coor = list(map(lambda x: x * scale, coor))

    return coor





'''

Functions used for calculating entropic measures (SMET Triangle)

'''

from sys import exit as exit
from warnings import warn as warning
from scipy.stats import entropy as entropy

from .auxfunc import * 


def sentropies(df, type = "total" , base = 2 , nbins = 1 ):

    """
    Source Entropy decomposition of a dataframe
    Given a dataframe, provide the dual or the aggregate entropy decomposition of the dataframe

    > edf = sentropies_df(df, type = 'total, base = 2, nbins = 1)

    Parameters
    ----------
    df : DataFrame for calculating the source entropies
    base : The logarithm to be used in working out the sentropies (Default base = 2 -> "log2")

    Returns
    ----------
    edf : Pandas DataFrame containing the values of the entropies calculated

    if the type = "total" then the function calculates the Total Decomposition & Aggregates
    else if  type = "dual" then it will calculate the dual Decomposition & Aggregates

    """

    dims = df.shape

    if(not isinstance(df,pd.DataFrame)):
        exit("Can only work with Data Frames!")

    if(dims[1] == 0 or dims[0] == 0): 
        exit("Can only work with non-empty DataFrames!")
    
    if (not(all(df.dtypes=='category'))):
        warning("Discretizing data from X DataFrame before entropy calculation!")
        df = discretization(df , nbins)

    H_Uxi = df.apply((lambda x : np.log2(len(pd.unique(x)))), axis = 0).values
    H_Pxi = df.apply(ent,axis=0).values

    if (dims[1]==1):

        warnings.warn("Single variable: providing only entropy")
        VI_Pxi = H_Pxi

    else:
        
        VI_Pxi = list()
        for i in range(len(df.columns)):
            VI_Pxi.append(condentropy(df[df.columns[i]],df[df.columns.drop(df.columns[i])]))
        
        
    if (type == "total") : # TOTAL decomposition & Aggregates

        if (any(df.columns.isnull())):
            #' Delete names and append new ones
            warning("No names for columns: providing variable names!")
            #Apply dummy names 
            df.columns = list(map(lambda x: "x"+str(x), range(len(df.columns))))

        #DataFrame Creation
        edf = pd.DataFrame({'Name': list(df.columns) , 'H_Uxi': H_Uxi , 'H_Pxi':H_Pxi ,'DeltaH_Pxi': H_Uxi - H_Pxi ,'M_Pxi': H_Pxi - VI_Pxi , 'VI_Pxi': VI_Pxi}, columns = ['Name','H_Uxi','H_Pxi','DeltaH_Pxi','M_Pxi','VI_Pxi'])
        edf = edf.set_index('Name')
        edf.loc['AGGREGATE'] = edf.sum(axis = 0)


    else :  #return only an aggregate with the DUAL total correlation D_Px

        H_Ux = np.sum(H_Uxi)
        H_Px = ent(sjoin(df,lis = df.columns))
        VI_Px = np.sum(VI_Pxi)

        edf = pd.DataFrame({'Name' : ['AGGREGATE'] , 'H_Ux' : H_Ux , 'H_Px'  : H_Px , 'DeltaH_Px' : (H_Ux - H_Px) ,'D_Px' : H_Px - VI_Px , 'VI_Px' : VI_Px }, columns = ['Name','H_Ux','H_Px','DeltaH_Px','D_Px','VI_Px'])
        edf = edf.set_index('Name')

    return edf


"""
Line plotting functions, draw boundary and gridlines.
"""

from numpy import arange
from matplotlib.lines import Line2D

from .helpers import project_point


## Lines ##

def line(ax, p1, p2, permutation=None, angle = None, **kwargs):
    """
    Draws a line on `ax` from p1 to p2.

    Parameters
    ----------
    ax: Matplotlib AxesSubplot, None
        The subplot to draw on.
    p1: 2-tuple
        The (x,y) starting coordinates
    p2: 2-tuple
        The (x,y) ending coordinates
    kwargs:
        Any kwargs to pass through to Matplotlib.
    """

    pp1 = project_point(p1, permutation=permutation, angle = angle)
    pp2 = project_point(p2, permutation=permutation, angle = angle)
    ax.add_line(Line2D((pp1[0], pp2[0]), (pp1[1], pp2[1]), **kwargs))


def horizontal_line(ax, scale, i, angle = None, **kwargs):
    """
    Draws the i-th horizontal line parallel to the lower axis.

    Parameters
    ----------
    ax: Matplotlib AxesSubplot
        The subplot to draw on.
    scale: float, 1.0
        Simplex scale size.
    i: float
        The index of the line to draw
    kwargs: Dictionary
        Any kwargs to pass through to Matplotlib.
    """

    p1 = (0, i, scale - i)
    p2 = (scale - i, i, 0)
    line(ax, p1, p2, angle = angle, **kwargs)


def left_parallel_line(ax, scale, i, angle = None,  **kwargs):
    """
    Draws the i-th line parallel to the left axis.

    Parameters
    ----------
    ax: Matplotlib AxesSubplot
        The subplot to draw on.
    scale: float
        Simplex scale size.
    i: float
        The index of the line to draw
    kwargs: Dictionary
        Any kwargs to pass through to Matplotlib.
    """

    p1 = (i, scale - i, 0)
    p2 = (i, 0, scale - i)
    line(ax, p1, p2, angle = angle, **kwargs)


def right_parallel_line(ax, scale, i, angle = None, **kwargs):
    """
    Draws the i-th line parallel to the right axis.

    Parameters
    ----------
    ax: Matplotlib AxesSubplot
        The subplot to draw on.
    scale: float
        Simplex scale size.
    i: float
        The index of the line to draw
    kwargs: Dictionary
        Any kwargs to pass through to Matplotlib.
    """

    p1 = (0, scale - i, i)
    p2 = (scale - i, 0, i)
    line(ax, p1, p2, angle = angle, **kwargs)


## Boundary, Gridlines ##

def boundary(ax, scale, axes_colors=None, angle = None, **kwargs):
    """
    Plots the boundary of the simplex. Creates and returns matplotlib axis if
    none given.

    Parameters
    ----------
    ax: Matplotlib AxesSubplot, None
        The subplot to draw on.
    scale: float
        Simplex scale size.
    kwargs:
        Any kwargs to pass through to matplotlib.
    axes_colors: dict
        Option for coloring boundaries different colors.
        e.g. {'l': 'g'} for coloring the left axis boundary green
    """

    # Set default color as black.
    if axes_colors is None:
        axes_colors = dict()
    for _axis in ['l', 'r', 'b']:
        if _axis not in axes_colors.keys():
            axes_colors[_axis] = 'black'

    horizontal_line(ax, scale, 0, angle = angle, color=axes_colors['b'], **kwargs)
    left_parallel_line(ax, scale, 0, angle = angle, color=axes_colors['l'], **kwargs)
    right_parallel_line(ax, scale, 0, angle = angle, color=axes_colors['r'], **kwargs)
    return ax


def merge_dicts(base, updates):
    """
    Given two dicts, merge them into a new dict as a shallow copy.

    Parameters
    ----------
    base: dict
        The base dictionary.
    updates: dict
        Secondary dictionary whose values override the base.
    """
    if not base:
        base = dict()
    if not updates:
        updates = dict()
    z = base.copy()
    z.update(updates)
    return z


def gridlines(ax, scale, multiple=None, horizontal_kwargs=None,
              left_kwargs=None, right_kwargs=None, angle= None, **kwargs):
    """
    Plots grid lines excluding boundary.

    Parameters
    ----------
    ax: Matplotlib AxesSubplot, None
        The subplot to draw on.
    scale: float
        Simplex scale size.
    multiple: float, None
        Specifies which inner gridelines to draw. For example, if scale=30 and
        multiple=6, only 5 inner gridlines will be drawn.
    horizontal_kwargs: dict, None
        Any kwargs to pass through to matplotlib for horizontal gridlines
    left_kwargs: dict, None
        Any kwargs to pass through to matplotlib for left parallel gridlines
    right_kwargs: dict, None
        Any kwargs to pass through to matplotlib for right parallel gridlines
    kwargs:
        Any kwargs to pass through to matplotlib, if not using
        horizontal_kwargs, left_kwargs, or right_kwargs
    """

    if 'linewidth' not in kwargs:
        kwargs["linewidth"] = 0.5
    if 'linestyle' not in kwargs:
        kwargs["linestyle"] = ':'
    horizontal_kwargs = merge_dicts(kwargs, horizontal_kwargs)
    left_kwargs = merge_dicts(kwargs, left_kwargs)
    right_kwargs = merge_dicts(kwargs, right_kwargs)
    if not multiple:
        multiple = 1.
    ## Draw grid-lines
    # Parallel to horizontal axis
    for i in arange(0, scale, multiple):
        horizontal_line(ax, scale, i, angle = angle, **horizontal_kwargs)
    # Parallel to left and right axes
    for i in arange(0, scale + multiple, multiple):
        left_parallel_line(ax, scale, i, angle = angle, **left_kwargs)
        right_parallel_line(ax, scale, i, angle = angle, **right_kwargs)
    return ax


def normalize_tick_formats(tick_formats):
    if type(tick_formats) == dict:
        return tick_formats
    if tick_formats is None:
        s = '%d'
    elif type(tick_formats) == str:
        s = tick_formats
    else:
        raise TypeError("tick_formats must be a dictionary of strings"
                        " a string, or None.")
    return {'b': s, 'l': s, 'r': s}


def ticks(ax, scale, ticks=None, locations=None, multiple=1, axis='b',
          offset=0.01, clockwise=False, axes_colors=None, fontsize=10,
          tick_formats=None, angle = None,  **kwargs):
    """
    Sets tick marks and labels.

    Parameters
    ----------
    ax: Matplotlib AxesSubplot, None
        The subplot to draw on.
    scale: float, 1.0
        Simplex scale size.
    ticks: list of strings, None
        The tick labels
    locations: list of points, None
        The locations of the ticks
    multiple: float, None
        Specifies which ticks gridelines to draw. For example, if scale=30 and
        multiple=6, only 5 ticks will be drawn.
    axis: str, 'b'
        The axis or axes to draw the ticks for. `axis` must be a substring of
        'lrb' (as sets)
    offset: float, 0.01
        controls the length of the ticks
    clockwise: bool, False
        Draw ticks marks clockwise or counterclockwise
    axes_colors: Dict, None
        Option to color ticks differently for each axis, 'l', 'r', 'b'
        e.g. {'l': 'g', 'r':'b', 'b': 'y'}
    tick_formats: None, Dict, Str
        If None, all axes will be labelled with ints. If Dict, the keys are
        'b', 'l' and 'r' and the values are format strings e.g. "%.3f" for
        a float with 3 decimal places or "%.3e" for scientific format with
        3 decimal places or "%d" for ints. If tick_formats is a string, it
        is assumed that this is a format string to be applied to all axes.
    kwargs:
        Any kwargs to pass through to matplotlib.

    """

    axis = axis.lower()
    valid_axis_chars = set(['l', 'r', 'b'])
    axis_chars = set(axis)
    if not axis_chars.issubset(valid_axis_chars):
        raise ValueError("axis must be some combination of 'l', 'r', and 'b'")

    if ticks and not locations:
        num_ticks = len(ticks)
        if num_ticks != 0:
            multiple = scale / (num_ticks - 1)
            locations = arange(0, scale + multiple, multiple)

    if not ticks:
        locations = arange(0, scale + multiple, multiple)
        ticks = locations

    tick_formats = normalize_tick_formats(tick_formats)

    # Default color: black
    if axes_colors is None:
        axes_colors = dict()
    for _axis in valid_axis_chars:
        if _axis not in axes_colors:
            axes_colors[_axis] = 'black'

    offset *= scale

    if 'r' in axis:
        for index, i in enumerate(locations):
            loc1 = (scale - i, i, 0)
            if clockwise:
                # Right parallel
                loc2 = (scale - i, i + offset, 0)
                text_location = (scale - i, i + 2 * offset, 0)
                tick = ticks[-(index+1)]
            else:
                # Horizontal
                loc2 = (scale - i + offset, i, 0)
                text_location = (scale - i + 2.6 * offset, i - 0.5 * offset, 0)
                tick = ticks[index]
            line(ax, loc1, loc2, angle = angle ,color=axes_colors['r'], **kwargs)
            x, y = project_point(text_location, angle = angle)
            s = tick_formats['r'] % tick
            ax.text(x, y, s, horizontalalignment="center",
                    color=axes_colors['r'], fontsize=fontsize)

    if 'l' in axis:
        for index, i in enumerate(locations):
            loc1 = (0, i, 0)
            if clockwise:
                # Horizontal
                loc2 = (-offset, i, 0)
                text_location = (-2 * offset, i - 0.5 * offset, 0)
                tick = ticks[index]
            else:
                # Right parallel
                loc2 = (-offset, i + offset, 0)
                text_location = (-2 * offset, i + 1.5 * offset, 0)
                tick = ticks[-(index+1)]
            line(ax, loc1, loc2, angle = angle,color=axes_colors['l'], **kwargs)
            x, y = project_point(text_location, angle = angle)
            s = tick_formats['l'] % tick
            ax.text(x, y, s, horizontalalignment="center",
                    color=axes_colors['l'], fontsize=fontsize)

    if 'b' in axis:
        for index, i in enumerate(locations):
            loc1 = (i, 0, 0)
            if clockwise:
                # Right parallel
                loc2 = (i + offset, -offset, 0)
                text_location = (i + 3 * offset, -3.5 * offset, 0)
                tick = ticks[-(index+1)]
            else:
                # Left parallel
                loc2 = (i, -offset, 0)
                text_location = (i + 0.5 * offset, -3.5 * offset, 0)
                tick = ticks[index]
            line(ax, loc1, loc2,angle = angle, color=axes_colors['b'], **kwargs)
            x, y = project_point(text_location, angle = angle)
            s = tick_formats['b'] % tick
            ax.text(x, y, s, horizontalalignment="center",
                    color=axes_colors['b'], fontsize=fontsize)



"""
Helper functions and utilities for projecting to the simplex and various tasks.
"""

import numpy


### Constants ###

SQRT3 = numpy.sqrt(3)
SQRT3OVER2 = SQRT3 / 2.

### Auxilliary Functions ###


def unzip(l):
    """[(a1, b1), ..., (an, bn)] ----> ([a1, ..., an], [b1, ..., bn])"""
    return list(zip(*l))


def normalize(l):
    """
    Normalizes input list.

    Parameters
    ----------
    l: list
        The list to be normalized

    Returns
    -------
    The normalized list or numpy array

    Raises
    ------
    ValueError, if the list sums to zero
    """

    s = float(sum(l))
    if s == 0:
        raise ValueError("Cannot normalize list with sum 0")
    return [x / s for x in l]


def simplex_iterator(scale, boundary=True):
    """
    Systematically iterates through a lattice of points on the 2-simplex.

    Parameters
    ----------
    scale: Int
        The normalized scale of the simplex, i.e. N such that points (x,y,z)
        satisify x + y + z == N

    boundary: bool, True
        Include the boundary points (tuples where at least one
        coordinate is zero)

    Yields
    ------
    3-tuples, There are binom(n+2, 2) points (the triangular
    number for scale + 1, less 3*(scale+1) if boundary=False
    """

    start = 0
    if not boundary:
        start = 1
    for i in range(start, scale + (1 - start)):
        for j in range(start, scale + (1 - start) - i):
            k = scale - i - j
            yield (i, j, k)


## Ternary Projections ##

## Ternary Projections ##

def rotate_point(p, angle=None):
    
        
    """
    Rotate the point p "angle" degrees, multiplying the point by the rotation matrix

    p = [x,y]

    rotation matrix
    | cos(angle) , -sin(angle) |
    | sin(angle) ,  cos(angle) |

    """
    if(isinstance(angle, type(None))):
        return p

    rotation_matrix = [[numpy.cos(numpy.deg2rad(angle)) , -numpy.sin(numpy.deg2rad(angle))] , [numpy.sin(numpy.deg2rad(angle)) , numpy.cos(numpy.deg2rad(angle))]]
    rotated_point = numpy.dot(rotation_matrix,p)
    return rotated_point

def permute_point(p, permutation=None):
    """
    Permutes the point according to the permutation keyword argument. The
    default permutation is "012" which does not change the order of the
    coordinate. To rotate counterclockwise, use "120" and to rotate clockwise
    use "201"."""
    if not permutation:
        return p
    return [p[int(permutation[i])] for i in range(len(p))]


def project_point(p, permutation=None, angle = None):
    """
    Maps (x,y,z) coordinates to planar simplex.

    Parameters
    ----------
    p: 3-tuple
        The point to be projected p = (x, y, z)
    permutation: string, None, equivalent to "012"
        The order of the coordinates, counterclockwise from the origin
    """
    permuted = permute_point(p, permutation=permutation)
    a = permuted[0]
    b = permuted[1]
    x = a + b/2.
    y = SQRT3OVER2 * b
    
    point = numpy.array([x,y])
    rotated_point = rotate_point(point, angle = angle)

    return rotated_point



def project_sequence(s, permutation=None, angle = None):
    """
    Projects a point or sequence of points using `project_point` to lists xs, ys
    for plotting with Matplotlib.

    Parameters
    ----------
    s, Sequence-like
        The sequence of points (3-tuples) to be projected.

    Returns
    -------
    xs, ys: The sequence of projected points in coordinates as two lists 
    """

    xs, ys = unzip([project_point(p, permutation=permutation, angle = angle) for p in s])
    return xs, ys


# Convert coordinates for custom plots with limits

def convert_coordinates(q, conversion, axisorder):
    """
    Convert a 3-tuple in data coordinates into to simplex data
    coordinates for plotting.

    Parameters
    ----------
    q: 3-tuple
       the point to be plotted in data coordinates

    conversion: dict
        keys = ['b','l','r']
        values = lambda function giving the conversion
    axisorder: String giving the order of the axes for the coordinate tuple
        e.g. 'blr' for bottom, left, right coordinates.

    Returns
    -------
    p: 3-tuple
       The point converted to simplex coordinates.
    """
    p = []
    for k in range(3):
        p.append(conversion[axisorder[k]](q[k]))

    return tuple(p)


def get_conversion(scale, limits):
    """
    Get the conversion equations for each axis.

    limits: dict of min and max values for the axes in the order blr.
    """
    fb = float(scale) / float(limits['b'][1] - limits['b'][0])
    fl = float(scale) / float(limits['l'][1] - limits['l'][0])
    fr = float(scale) / float(limits['r'][1] - limits['r'][0])

    conversion = {"b": lambda x: (x - limits['b'][0]) * fb,
                  "l": lambda x: (x - limits['l'][0]) * fl,
                  "r": lambda x: (x - limits['r'][0]) * fr}

    return conversion
    

def convert_coordinates_sequence(qs, scale, limits, axisorder):
    """
    Take a sequence of 3-tuples in data coordinates and convert them
    to simplex coordinates for plotting. This is needed for custom
    plots where the scale of the simplex axes is set within limits rather
    than being defined by the scale parameter.

    Parameters
    ----------
    qs, sequence of 3-tuples
       The points to be plotted in data coordinates.

    scale: int
        The scale parameter for the plot.
    limits: dict
        keys = ['b','l','r']
        values = min,max data values for this axis.
    axisorder: String giving the order of the axes for the coordinate tuple
        e.g. 'blr' for bottom, left, right coordinates.

    Returns
    -------
    s, list of 3-tuples
       the points converted to simplex coordinates
    """
    conversion = get_conversion(scale, limits)
    
    return [convert_coordinates(q, conversion, axisorder) for q in qs]



from setuptools import setup,find_packages

classifiers = [
    "Development Status :: 3 - Alpha",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: MIT License",
    "Natural Language :: English",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.6",
]

with open('README.txt') as file:
    long_description = file.read()
    
with open('requirements.txt') as reqs:
	install_requires = reqs.read().splitlines()

print(find_packages())
setup(
    name="entropytriangle",
    version="1.0.1",
    packages= find_packages(),
    python_requires='>=3',
    install_requires = install_requires,
    author="Jaime de los Rios Mouvet",
    author_email="jaime.delosriosmouvet@gmail.com",
    classifiers=classifiers,
    description="Calculation of the entropy triangles",
    long_description=long_description,
    keywords="Entropy Triangle Information Theory",
    license="MIT",
    url="https://github.com/Jaimedlrm/entropytriangle",
    download_url="https://github.com/Jaimedlrm/entropytriangle.git",
)




import matplotlib
from matplotlib import pyplot as plt
from matplotlib.colors import rgb2hex

## Default colormap, other options here: http://www.scipy.org/Cookbook/Matplotlib/Show_colormaps
s = matplotlib.__version__.split('.')
if int(s[0]) >= 2 or (int(s[0]) >= 1 and int(s[1]) >= 5):
    DEFAULT_COLOR_MAP_NAME = "viridis"
else:
    DEFAULT_COLOR_MAP_NAME = "jet"


## Matplotlib Colormapping ##

def get_cmap(cmap=None):
    """
    Loads a matplotlib colormap if specified or supplies the default.

    Parameters
    ----------
    cmap: string or matplotlib.colors.Colormap instance
        The name of the Matplotlib colormap to look up.

    Returns
    -------
    The desired Matplotlib colormap

    Raises
    ------
    ValueError if colormap name is not recognized by Matplotlib
    """

    if isinstance(cmap, matplotlib.colors.Colormap):
        return cmap
    if isinstance(cmap, str):
        cmap_name = cmap
    else:
        cmap_name = DEFAULT_COLOR_MAP_NAME
    return plt.get_cmap(cmap_name)


def colormapper(value, lower=0, upper=1, cmap=None):
    """
    Maps values to colors by normalizing within [a,b], obtaining rgba from the
    given matplotlib color map for heatmap polygon coloring.

    Parameters
    ----------
    value: float
        The value to be colormapped
    lower: float
        Lower bound of colors
    upper: float
        Upper bound of colors
    cmap: String or matplotlib.colors.Colormap (optional)
        Colormap object to prevent repeated lookup

    Returns
    -------
    hex_, float
        The value mapped to an appropriate RGBA color value
    """

    cmap = get_cmap(cmap)
    if upper - lower == 0:
        rgba = cmap(0)
    else:
        rgba = cmap((value - lower) / float(upper - lower))
    hex_ = rgb2hex(rgba)
    return hex_


def colorbar_hack(ax, vmin, vmax, cmap, scientific=False, cbarlabel=None,
                  **kwargs):
    """
    Colorbar hack to insert colorbar on ternary plot. 
    
    Called by heatmap, not intended for direct usage.
    
    Parameters
    ----------
    vmin: float
        Minimum value to portray in colorbar
    vmax: float
        Maximum value to portray in colorbar
    cmap: Matplotlib colormap
        Matplotlib colormap to use

    """
    # http://stackoverflow.com/questions/8342549/matplotlib-add-colorbar-to-a-sequence-of-line-plots
    norm = plt.Normalize(vmin=vmin, vmax=vmax)
    sm = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
    sm._A = []
    cb = plt.colorbar(sm, ax=ax, **kwargs)
    if cbarlabel is not None:
        cb.set_label(cbarlabel)
    if scientific:
        cb.locator = matplotlib.ticker.LinearLocator(numticks=7)
        cb.formatter = matplotlib.ticker.ScalarFormatter()
        cb.formatter.set_powerlimits((0, 0))
        cb.update_ticks()
